#include "mpi.h"
#include "mpe.h"
#include <stdio.h>
#include <math.h>

extern void   srand48();
extern double drand48();

/* we really need the velocities as well... */
typedef struct {
    double x, y, z;
    double mass;
    } Particle;
/* We use leapfrog for the time integration ... */
typedef struct {
    double xold, yold, zold;
    } ParticleV;

#define MAX_PARTICLES 4000
#define MAX_P          128
main( argc, argv )
int  argc;
char **argv;
{
Particle  particles[MAX_PARTICLES];   /* Particles on ALL nodes */
ParticleV pv[MAX_PARTICLES];          /* Particle velocity */
int       counts[MAX_P],              /* Number on each processor */
          displs[MAX_P];              /* Offsets into particles */
int       rank, size, npart, i, j,
          offset;                     /* location of local particles */
int       totpart,                    /* total number of particles */
          cnt;                        /* number of times in loop */
MPI_Datatype particletype;
double    time;                       /* Computation time */
double    dt, dt_old;                 /* Integration time step */
double    a0, a1, a2;

MPI_Init( &argc, &argv );
MPI_Comm_rank( MPI_COMM_WORLD, &rank );
MPI_Comm_size( MPI_COMM_WORLD, &size );
/* Everyone COULD have a different size ... */
npart = atoi(argv[1]) / size;

if (npart * size > MAX_PARTICLES) 
    MPI_Abort( MPI_COMM_WORLD, 1 );

MPI_Type_contiguous( 4, MPI_DOUBLE, &particletype );
MPI_Type_commit( &particletype );

/* Get the sizes and displacements */
MPI_Allgather( &npart, 1, MPI_INT, counts, 1, MPI_INT, MPI_COMM_WORLD );
displs[0] = 0;
for (i=1; i<size; i++) 
    displs[i] = displs[i-1] + counts[i-1];
totpart = displs[size-1] + counts[size-1];

/* Generate the initial values */
offset = displs[rank];
for (i=0; i<npart; i++) {
    particles[offset+i].x    = drand48();
    particles[offset+i].y    = drand48();
    particles[offset+i].z    = drand48();
    particles[offset+i].mass = 1.0;
    pv[i].xold		     = particles[offset+i].x;
    pv[i].yold		     = particles[offset+i].y;
    pv[i].zold		     = particles[offset+i].z;
    }
cnt    = 10;
dt     = 0.001;
dt_old = 0.001;
/* integation is a0 * x^+ + a1 * x + a2 * x^- = f / m */
a0   = 2.0 / (dt * (dt + dt_old));
a2   = 2.0 / (dt_old * (dt + dt_old));
a1   = -(a0 + a1);      /* also -2/(dt*dt_old) */
time = MPI_Wtime();
while (cnt--) {
    double max_f, dt_est, new_dt, dt_new;
    MPI_Allgatherv( particles+offset, npart, particletype, particles,
		    counts, displs, particletype, MPI_COMM_WORLD );
    /* Compute forces (2D only) */
    max_f = 0.0;
    for (i=0; i<npart; i++) {
	double xi, yi, mi, rx, ry, mj, r, fx, fy;
	double xnew, ynew, rmin;
	rmin = 100.0;
	xi   = particles[offset+i].x;
	yi   = particles[offset+i].y;
	fx   = 0.0;
	fy   = 0.0;
	for (j=0; j<totpart; j++) {
	    rx = xi - particles[j].x;
	    ry = yi - particles[j].y;
	    mj = particles[j].mass;
	    r  = rx * rx + ry * ry;
		/* ignore overlap and same particle */
	    if (r == 0.0) continue;
	    if (r < rmin) rmin = r;
	    /* compute forces */
	    r  = r * sqrt(r);
	    fx += mj * rx / r;
	    fy += mj * ry / r;
	    }
	/* Very, very simple leapfrog time integration.  We use a variable 
	   step version to simplify time-step control. */
	particles[offset+i].x = (fx - a1 * xi - a2 * pv[i].xold) / a0;
	particles[offset+i].y = (fy - a1 * yi - a2 * pv[i].yold) / a0;
	pv[i].xold	      = xi;
	pv[i].yold	      = yi;
	/* Compute a rough estimate of (1/m)|df / dx| */
	fx		      = sqrt(fx*fx + fy*fy)/rmin;
	if (fx > max_f) max_f = fx;
	}
    /* Recompute a time step. Stability criteria is roughly 
       sqrt(1/m |df/dx|)/2 >= dt.  We leave a little room */
    dt_est = 0.25*sqrt(max_f);
    /* Set a minimum: */
    if (dt_est < 1.0e-6) dt_est = 1.0e-6;
    MPI_Allreduce( &dt_est, &dt_new, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD );
    /* Modify time step */
    if (dt_new < dt) {
	dt_old = dt;
	dt     = dt_new;
	}
    else if (dt_new > 4.0 * dt) {
	dt_old = dt;
	dt    *= 2.0;
	}
    /* Recompute the time step integrator in case we changed the step sizes */
       a0   = 2.0 / (dt * (dt + dt_old));
       a2   = 2.0 / (dt_old * (dt + dt_old));
       a1   = -(a0 + a1);      /* also -2/(dt*dt_old) */
    /* We could do graphics here (move particles on the display) */
    }
time = MPI_Wtime() - time;
if (rank == 0) {
    printf( "Computed %d particles in %f seconds\n", totpart, time );
    }
MPI_Finalize();
}
