#include "mpi.h"
#include "mpe.h"
#include <stdio.h>
#include <math.h>

extern void   srand48();
extern double drand48();

/* 
   G is the gravitational constant.  I need to look it up..
   It also needs to be scaled to the units used here (mass, distance, time) 
 */
#define G 1.0e-6

/* Pipeline version of the algorithm, using persistant requests ... */
/* we really need the velocities as well... */
typedef struct {
    double x, y, z;
    double mass;
    } Particle;
/* We use leapfrog for the time integration ... */
typedef struct {
    double xold, yold, zold;
    double fx, fy, fz;
    } ParticleV;

#define MAX_PARTICLES 4000
#define MAX_P          128
main( argc, argv )
int  argc;
char **argv;
{
Particle  particles[MAX_PARTICLES];   /* Particles on ALL nodes */
ParticleV pv[MAX_PARTICLES];          /* Particle velocity */
Particle  sendbuf[MAX_PARTICLES],     /* Pipeline buffers */
          recvbuf[MAX_PARTICLES];
MPI_Request request[2*MAX_P];
int       counts[MAX_P],              /* Number on each processor */
          displs[MAX_P];              /* Offsets into particles */
int       rank, size, npart, i, j,
          offset;                     /* location of local particles */
int       totpart,                    /* total number of particles */
          cnt;                        /* number of times in loop */
MPI_Datatype particletype;
double    time;                       /* Computation time */
double    dt, dt_old;                 /* Integration time step */
double    a0, a1, a2;
int       pipe, left, right, periodic;
MPI_Comm  commring;
MPI_Status statuses[2];

MPI_Init( &argc, &argv );
MPI_Comm_rank( MPI_COMM_WORLD, &rank );
MPI_Comm_size( MPI_COMM_WORLD, &size );

/* Get the best ring in the topology */
periodic = 1;
MPI_Cart_create( MPI_COMM_WORLD, 1, &size, &periodic, 1, &commring );
MPI_Cart_shift( commring, 0, 1, &left, &right );

SYSetFPTraps();
/* Everyone COULD have a different size ... */
npart = atoi(argv[1]) / size;

if (npart * size > MAX_PARTICLES) 
    MPI_Abort( MPI_COMM_WORLD, 1 );

MPI_Type_contiguous( 4, MPI_DOUBLE, &particletype );
MPI_Type_commit( &particletype );

/* Get the sizes and displacements */
MPI_Allgather( &npart, 1, MPI_INT, counts, 1, MPI_INT, commring );
displs[0] = 0;
for (i=1; i<size; i++) 
    displs[i] = displs[i-1] + counts[i-1];
totpart = displs[size-1] + counts[size-1];

/* Generate the requests */
/* This requires that the ranks line up with the natural ordering... */
for (i=0; i<size-1; i++) {
    MPI_Send_init( sendbuf, counts[(rank+i)%size], 
		   particletype, right, i, commring, 
	             &request[2*i] );
    MPI_Recv_init( recvbuf, counts[(rank+i-1+size)%size], particletype, 
	           left, i, commring, &request[2*i+1] );
    }

/* Generate the initial values */
offset = displs[rank];
for (i=0; i<npart; i++) {
    particles[i].x    = drand48();
    particles[i].y    = drand48();
    particles[i].z    = drand48();
    particles[i].mass = 1.0;
    pv[i].xold	      = particles[i].x;
    pv[i].yold	      = particles[i].y;
    pv[i].zold	      = particles[i].z;
    pv[i].fx	      = 0;
    pv[i].fy	      = 0;
    pv[i].fz	      = 0;
    }
cnt    = 10;
dt     = 0.001;
dt_old = 0.001;
/* integation is a0 * x^+ + a1 * x + a2 * x^- = f / m */
a0   = 2.0 / (dt * (dt + dt_old));
a2   = 2.0 / (dt_old * (dt + dt_old));
a1   = -(a0 + a1);      /* also -2/(dt*dt_old) */
time = MPI_Wtime();
while (cnt--) {
    double max_f, dt_est, new_dt, dt_new;
    
    /* Load the initial sendbuffer */
    memcpy( sendbuf, particles, npart * sizeof(Particle) );
    for (pipe=0; pipe<size; pipe++) {
	if (pipe != size-1) 
	    MPI_Startall( 2, &request[2*pipe] );
	/* Compute forces (2D only) */
	max_f = 0.0;
	for (i=0; i<npart; i++) {
	    double xi, yi, mi, rx, ry, mj, r, fx, fy;
	    double xnew, ynew, rmin;
	    rmin = 100.0;
	    xi   = particles[i].x;
	    yi   = particles[i].y;
	    fx   = 0.0;
	    fy   = 0.0;
	    for (j=0; j<npart; j++) {
		rx = xi - sendbuf[j].x;
		ry = yi - sendbuf[j].y;
		mj = sendbuf[j].mass;
		r  = rx * rx + ry * ry;
		/* ignore overlap and same particle */
		if (r == 0.0) continue;
		if (r < rmin) rmin = r;
		/* compute forces */
		r  = r * sqrt(r);
		fx += mj * rx / r;
		fy += mj * ry / r;
		}
	    pv[i].fx += fx;
	    pv[i].fy += fy;
	    /* Compute a rough estimate of (1/m)|df / dx| */
	    fx		      = sqrt(fx*fx + fy*fy)/rmin;
	    if (fx > max_f) max_f = fx;
	    }
	/* Push pipe */
	if (pipe != size-1) 
	    MPI_Waitall( 2, &request[2*pipe], statuses );
	memcpy( sendbuf, recvbuf, npart * sizeof(Particle) );
	}
    /* Once we have the forces, we compute the changes in position */
    for (i=0; i<npart; i++) {
	double xi, yi;
	/* Very, very simple leapfrog time integration.  We use a variable 
	   step version to simplify time-step control. */
	xi	       = particles[i].x;
	yi	       = particles[i].y;
	particles[i].x = (G*pv[i].fx - a1 * xi - a2 * pv[i].xold) / a0;
	particles[i].y = (G*pv[i].fy - a1 * yi - a2 * pv[i].yold) / a0;
	pv[i].xold     = xi;
	pv[i].yold     = yi;
	pv[i].fx       = 0;
	pv[i].fy       = 0;
	}
    /* Recompute a time step. Stability criteria is roughly 
       sqrt(1/m |df/dx|)/2 >= dt.  We leave a little room */
    dt_est = 0.25*sqrt(max_f);
    /* Set a minimum: */
    if (dt_est < 1.0e-6) dt_est = 1.0e-6;
    MPI_Allreduce( &dt_est, &dt_new, 1, MPI_DOUBLE, MPI_MIN, commring );
    /* Modify time step */
    if (dt_new < dt) {
	dt_old = dt;
	dt     = dt_new;
	}
    else if (dt_new > 4.0 * dt) {
	dt_old = dt;
	dt    *= 2.0;
	}
    /* Recompute the time step integrator in case we changed the step sizes */
       a0   = 2.0 / (dt * (dt + dt_old));
       a2   = 2.0 / (dt_old * (dt + dt_old));
       a1   = -(a0 + a1);      /* also -2/(dt*dt_old) */
    /* We could do graphics here (move particles on the display) */
    }
time = MPI_Wtime() - time;
if (rank == 0) {
    printf( "Computed %d particles in %f seconds\n", totpart, time );
    }
for (i=0; i<2*(size-1); i++) {
    MPI_Request_free( &request[i] );
    }

MPI_Finalize();
}
